/*
 * bqfs.c
 * Copyright (C) 2002-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <libx1f4l2.h>

#define __UNROLL_PICK__			0

#define __SIMPLE_PICK__			0

#define __SKIP_LLZD_TEST__		1
#define __SKIP_LXZD_TEST__		1

#define __PICK_TAIL__			1

#if __SKIP_LLZD_TEST__
# if __SKIP_LXZD_TEST__
#  undef __PICK_TAIL__
#  define __PICK_TAIL__			0
# endif				/* __SKIP_LXZD_TEST__ */
#endif				/* __SKIP_LLZD_TEST__ */

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_VOID_P == 8
# define HALF				040
#else
# define HALF				020
#endif				/* SIZEOF_VOID_P == 8 */

#define VX				65536

typedef struct context_type {
    const char *self;
    int do_explain;
    unsigned count, size;
} context_type;

typedef struct test_type {
    const char *name;
    int (*test) (struct context_type *, void *, void *);
} test_type;

static int fine_look(void *, void *, unsigned *);
static int fine_fare(void *, void *);
static int fine_move(void *, void *, unsigned);
static int fine_pick(void *, void *, unsigned *);
static int free_list(void *);
static int line_list(struct context_type *, void **);
static int pick_head(void *, void *);
#if __PICK_TAIL__
static int pick_tail(void *, void *);
#endif				/* __PICK_TAIL__ */
static int post_list(struct context_type *, void *, void *);
static int test_fgxd(struct context_type *, void *, void *);
static int test_fgzd(struct context_type *, void *, void *);
static int test_fmxd(struct context_type *, void *, void *);
static int test_fxxd(struct context_type *, void *, void *);
static int test_fxzd(struct context_type *, void *, void *);
static int test_llxd(struct context_type *, void *, void *);
#if __SKIP_LLZD_TEST__
#else
static int test_llzd(struct context_type *, void *, void *);
#endif				/* __SKIP_LLZD_TEST__ */
static int test_lmxd(struct context_type *, void *, void *);
static int test_lxxd(struct context_type *, void *, void *);
#if __SKIP_LXZD_TEST__
#else
static int test_lxzd(struct context_type *, void *, void *);
#endif				/* __SKIP_LXZD_TEST__ */

static const struct test_type tests[] = {
/* *INDENT-OFF* */
    {   "last less search",
	test_llxd				},
    {   "last less or matching search",
	test_lxxd				},
    {   "last matching search",
	test_lmxd				},
    {   "first matching search",
	test_fmxd				},
    {   "first matching or greater search",
	test_fxxd				},
    {   "first greater search",
	test_fgxd				},
#if __SKIP_LLZD_TEST__
#else
    {   "lesser traverse",
	test_llzd				},
#endif				/* __SKIP_LLZD_TEST__ */
#if __SKIP_LXZD_TEST__
#else
    {   "lesser or matching traverse",
	test_lxzd				},
#endif				/* __SKIP_LXZD_TEST__ */
    {   "matching or greater traverse",
	test_fxzd				},
    {   "greater traverse",
	test_fgzd				},
    {   "",
	NULL					}
/* *INDENT-ON* */
};

static void usage(void);

static int
fine_fare(void *text, void *node)
{
    return *(unsigned long *) text < *(unsigned long *) node
	? -1 : *(unsigned long *) text ^ *(unsigned long *) node ? 1 : 0;
}


static int
fine_look(void *text, void *node, unsigned *pick)
{
    int status;
    unsigned size;

    size = *(integral_q *) node;
    if (size) {
	unsigned long *call, fine, *high;

	fine = *(unsigned long *) text;

	high = node;

	call = high + 1;

	size--;
	while (size) {
	    unsigned half;
	    unsigned long *slip;

	    half = (size + 1) >> 1;
	    slip = call + half;
	    if (fine < *slip) {
		size = half - 1;
	    } else {
		call = slip;
		size -= half;
	    }
	}

	if (*call ^ fine) {
	    status = 0;

	    if (fine < *call) {
		call--;
	    } else {
	    }

	    *pick = call - high;
	} else {
	    status = 1;

	    *pick = call - high;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
}


static int
fine_move(void *target, void *source, unsigned count)
{
#if 1
    unsigned long *s, *t;

    s = source;
    t = target;
    if (count ^ 1) {
	if (t < s) {
	    for (; count; count--) {
		*t++ = *s++;
	    }
	} else {
	    s += count;
	    if (t < s) {
		t += count;
		for (; count; count--) {
		    *--t = *--s;
		}
	    } else {
		s -= count;
		for (; count; count--) {
		    *t++ = *s++;
		}
	    }
	}
    } else {
	*t++ = *s++;
    }
#else
    memmove(target, source, count * sizeof(unsigned long));
#endif				/* 0 */

    return 0;
}


static int
fine_pick(void *text, void *node, unsigned *pick)
{
    int status;
#if __UNROLL_PICK__
#else
    unsigned half = HALF;
#endif				/* __UNROLL_PICK__ */
    unsigned slip = 0;
    unsigned long fine, *high;

    fine = *(unsigned long *) text;

    high = node;

#if __UNROLL_PICK__
# if SIZEOF_VOID_P == 8
    if (fine < high[slip + 0040]) {
    } else {
	slip += 0040;
    }
# endif				/* SIZEOF_VOID_P == 8 */
    if (fine < high[slip + 0020]) {
    } else {
	slip += 0020;
    }
    if (fine < high[slip + 0010]) {
    } else {
	slip += 0010;
    }
    if (fine < high[slip + 0004]) {
    } else {
	slip += 0004;
    }
    if (fine < high[slip + 0002]) {
    } else {
	slip += 0002;
    }
#else
    do {
	if (fine < high[slip + half]) {
	} else {
	    slip += half;
	}

	half >>= 1;
    } while (half ^ 1);
#endif				/* __UNROLL_PICK__ */

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	if (fine < high[slip + 1]) {
	    *pick = slip + 0;
	    if (slip) {
		if (fine ^ high[slip + 0]) {
		    status = 0;
		} else {
		    status = 1;
		}
	    } else {
		status = 0;
	    }
	} else {
	    *pick = slip + 1;
	    if (fine ^ high[slip + 1]) {
		status = 0;
	    } else {
		status = 1;
	    }
	}
    } else {
	*pick = slip + 1;
	if (slip) {
	    if (fine ^ high[slip + 0]) {
		status = 0;
	    } else {
		status = 1;
	    }
	} else {
	    if (1) {
		status = 0;
	    }
	}
    }

    return status;
}


static int
free_list(void *list)
{
    free(list);

    return 0;
}


static int
line_list(struct context_type *context_data, void **link)
{
    int status;
    void *list;

    list = malloc(sizeof(unsigned long) * context_data->count);
    if (list) {
	unsigned i, v = 1;
	unsigned long *line;

	status = 0;

	*link = list;

	line = list;

	i = context_data->count;
	while (i) {
	    unsigned e;

	    e = i * 523 + 342;

	    v *= 1033;

	    i--;

	    e ^= v;

#if SIZEOF_UNSIGNED != 4
	    e &= 0xffffffffu;
#endif				/* SIZEOF_UNSIGNED != 4 */

	    e = e >> 16 | e << 16;

	    *line++ = e;
	}
    } else {
	status = -1;
    }

    return status;
}


static int
pick_head(void *copy, void *head)
{
    void **data;

    data = copy;
    *data = head;

    return 1;
}


#if __PICK_TAIL__
static int
pick_tail(void *copy, void *head)
{
    void **data;

    data = copy;
    *data = head;

    return 0;
}
#endif				/* __PICK_TAIL__ */


static int
post_list(struct context_type *context_data, void *bqfx, void *list)
{
    int shift, status;
    unsigned i;

    shift = context_data->size;

    i = context_data->count;
    while (i) {
	void *post;

	i--;
	status = x1f4_post_bqfset(bqfx, list, &post);
	if (status) {
	    fprintf(stderr, "%s: cannot insert data\n", context_data->self);
	    if (1) {
		break;
	    }
	} else {
	    memcpy(post, list, shift);

	    list = (char *) list + shift;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
test_fgxd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	unsigned long seek;
	void *data;

	seek = *line;

	i--;

	if (seek) {
	    seek--;

	    status = x1f4_fgxd_bqfset(bqfx, &seek, &data);
	    if (status) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
test_fgzd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	unsigned long seek;
	void *data;

	seek = *line;

	i--;

	if (seek) {
	    seek--;

	    status = x1f4_fgzd_bqfset(bqfx, &seek, &data, pick_head);
	    if (status) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
test_fmxd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	void *data;

	i--;

	if (1) {
	    status = x1f4_fmxd_bqfset(bqfx, line, &data);
	    if (status) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
test_fxxd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	unsigned long seek;
	void *data;

	seek = *line;

	i--;

	if (seek) {
	    seek--;

	    status = x1f4_fxxd_bqfset(bqfx, &seek, &data);
	    if (status) {
		unsigned long find;

		find = *(unsigned long *) data;
		if (find ^ *line && find ^ seek) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
test_fxzd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	void *data;

	i--;

	if (1) {
	    status = x1f4_fxzd_bqfset(bqfx, line, &data, pick_head);
	    if (status) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
test_llxd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	unsigned long seek;
	void *data;

	seek = *line;

	i--;

	if (~seek) {
	    seek++;

	    status = x1f4_llxd_bqfset(bqfx, &seek, &data);
	    if (status) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


#if __SKIP_LLZD_TEST__
#else
static int
test_llzd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	unsigned long seek;
	void *data;

	seek = *line;

	i--;

	if (~seek) {
	    data = NULL;

	    seek++;

	    x1f4_llzd_bqfset(bqfx, &seek, &data, pick_tail);
	    if (data) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __SKIP_LLZD_TEST__ */


static int
test_lmxd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	void *data;

	i--;

	if (1) {
	    status = x1f4_lmxd_bqfset(bqfx, line, &data);
	    if (status) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


static int
test_lxxd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	unsigned long seek;
	void *data;

	seek = *line;

	i--;

	if (~seek) {
	    seek++;

	    status = x1f4_lxxd_bqfset(bqfx, &seek, &data);
	    if (status) {
		unsigned long find;

		find = *(unsigned long *) data;
		if (find ^ *line && find ^ seek) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}


#if __SKIP_LXZD_TEST__
#else
static int
test_lxzd(struct context_type *context_data, void *bqfx, void *list)
{
    int status;
    unsigned i;
    unsigned long *line;

    line = list;

    i = context_data->count;
    while (i) {
	void *data;

	i--;

	if (1) {
	    data = NULL;

	    x1f4_lxzd_bqfset(bqfx, line, &data, pick_tail);
	    if (data) {
		if (*(unsigned long *) data ^ *line) {
		    fprintf(stderr, "%s: bad pick, looking for %u, got %u\n",
			    context_data->self, (unsigned) *line,
			    (unsigned) *(unsigned long *) data);
		    if (1) {
			break;
		    }
		} else {
		    line++;
		}
	    } else {
		fprintf(stderr, "%s: cannot find %u\n", context_data->self,
			(unsigned) *line);
		if (1) {
		    break;
		}
	    }
	} else {
	    line++;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __SKIP_LXZD_TEST__ */


static void
usage(void)
{
    puts("Usage: bqfs [OPTIONS]\n\
Test the BQ generic set search methods.\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -v, --verbose			explain what is being done\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status;
    struct context_type context;

    context.do_explain = 0;
    context.count = VX;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *size_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "verbose",	    0x00,   NULL,   0x76   },
		{   "version",	    0x00,   NULL,   0x56    },
		{   "help",	    0x00,   NULL,   0x48    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:v", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'v':
		context.do_explain = 1;
		break;
	    case 'H':
	    case 'V':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'N':
		size_specs = optarg;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'H':
		usage();
		break;
	    case 'V':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse test set width specificati"
			"on: `%s'\n", argv[0], size_specs);

		return 1;
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct x1f4_bqfset_type bqfset;
	void *bqfx;

	if (0) {
	} else {
	    context.size = sizeof(unsigned long);
	}

	if (0) {
	} else {
	    bqfset.size = sizeof(unsigned long);
	}

	if (0) {
	} else {
	    bqfset.fare = fine_fare;
	    bqfset.look = fine_look;
	    bqfset.move = fine_move;
	    bqfset.pick = fine_pick;
	}

	status = x1f4_init_bqfset(&bqfx, X1f4_BQFSET_SIZE_FRAME, &bqfset);
	if (status) {
	    perror(argv[0]);
	} else {
	    int excess;
	    void *list;

	    context.self = argv[0];

	    if (0) {
	    } else {
		status = line_list(&context, &list);
	    }
	    if (status) {
		perror(argv[0]);
	    } else {
		status = post_list(&context, bqfx, list);
		if (status) {
		} else {
		    int (*test) (struct context_type *, void *, void *);
		    const struct test_type *test_data;

		    test_data = tests;

		    test = test_data->test;
		    while (test) {
			if (context.do_explain) {
			    fprintf(stderr, "running %s test...\n",
				    test_data->name);
			}

			status = test(&context, bqfx, list);
			if (status) {
			    break;
			} else {
			    test_data++;
			    test = test_data->test;
			}
		    }
		}

		free_list(list);
	    }

	    excess = x1f4_fini_bqfset(&bqfx);
	    if (excess) {
		perror(argv[0]);

		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}
