/*
 * bqpx.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <count2.h>
#include <libx1f4l2.h>

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_VOID_P == 8
# define HALF				040
#else
# define HALF				020
#endif				/* SIZEOF_VOID_P == 8 */

#define VX				65536

typedef struct context_type {
    const char *self;
    int do_profile, do_reverse, do_sort;
    unsigned count;
} context_type;

static int cast_list(struct context_type *, void *, void *);
static int find_list(struct context_type *, void *, void *);
static int free_list(void *);
static int line_list(struct context_type *, void **);
static int post_list(struct context_type *, void *, void *);
static int sort_fare(const void *, const void *);

static void usage(void);

static const char digits[] = "0123456789";

static int
cast_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int shift, status;
    unsigned i;

    shift = sizeof(integral_q);

    if (context_data->do_sort) {
	qsort(list, context_data->count, shift, sort_fare);
    }

    if (context_data->do_reverse) {
	list = (char *) list + (context_data->count - 1) * shift;
	shift = -shift;
    }

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	i--;
	status = x1f4_cast_bqpset(bqfx, *(void **) list);
	if (status) {
	    fprintf(stderr, "%s: cannot delete data\n", context_data->self);
	    if (1) {
		break;
	    }
	} else {
	    list = (char *) list + shift;
	}
    }

    if (i) {
	status = 1;
    } else {
	if (x1f4_case_bqfset(bqfx)) {
	    fprintf(stderr, "%s: set not void\n", context_data->self);

	    status = 1;
	} else {
	    status = 0;

	    if (context_data->do_profile) {
		finish_count2_l("delete", 55);
	    }
	}
    }

    return status;
}


static int
find_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int status;
    unsigned i;

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	i--;
	status = x1f4_find_bqpset(bqfx, *(void **) list);
	if (status) {
	    list = (char *) list + sizeof(integral_q);
	} else {
	    fprintf(stderr, "%s: cannot find data\n", context_data->self);
	    if (1) {
		break;
	    }
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;

	if (context_data->do_profile) {
	    finish_count2_l("seek", 55);
	}
    }

    return status;
}


static int
free_list(void *list)
{
    free(list);

    return 0;
}


static int
line_list(struct context_type *context_data, void **link)
{
    int status;
    void *list;

    list = malloc(sizeof(integral_q) * context_data->count);
    if (list) {
	integral_q *line;
	unsigned i, v = 1;

	status = 0;

	*link = list;

	line = list;

	i = context_data->count;
	while (i) {
	    unsigned e;

	    e = i * 523 + 342;

	    v *= 1033;

	    i--;

	    e ^= v;

#if SIZEOF_UNSIGNED != 4
	    e &= 0xffffffffu;
#endif				/* SIZEOF_UNSIGNED != 4 */

	    e = e >> 16 | e << 16;

	    *line++ = e;
	}
    } else {
	status = -1;
    }

    return status;
}


static int
post_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int status;
    unsigned i;

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	i--;
	status = x1f4_post_bqpset(bqfx, *(void **) list);
	if (status) {
	    fprintf(stderr, "%s: cannot insert data\n", context_data->self);
	    if (1) {
		break;
	    }
	} else {
	    list = (char *) list + sizeof(integral_q);
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;

	if (context_data->do_profile) {
	    finish_count2_l("insert", 55);
	}
    }

    return status;
}


static int
sort_fare(const void *text, const void *node)
{
    return *(integral_q *) text < *(integral_q *) node
	? -1 : *(integral_q *) text ^ *(integral_q *) node ? 1 : 0;
}


static void
usage(void)
{
    puts("Usage: bqpx [OPTIONS]\n\
Test the BQ pointer set.\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -p, --profile			time operations\n\
  -r, --reverse			remove items in reversed order\n\
  -s, --sort			remove items in sort order\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status;
    struct context_type context;

    context.do_profile = 0;
    context.do_reverse = 0;

    context.do_sort = 0;

    context.count = VX;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *size_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "sort",	    0x00,   NULL,   0x73    },
		{   "profile",	    0x00,   NULL,   0x70    },
		{   "reverse",	    0x00,   NULL,   0x72    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:prs", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'p':
		context.do_profile = 1;
		break;
	    case 'r':
		context.do_reverse = 1;
		break;
	    case 's':
		context.do_sort = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse test set width specificati"
			"on: `%s'\n", argv[0], size_specs);

		return 1;
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct x1f4_bqfset_type bqfset;
	void *bqfx;

	status = x1f4_init_bqfset(&bqfx, 0, &bqfset);
	if (status) {
	    perror(argv[0]);
	} else {
	    int excess;
	    void *list;

	    context.self = argv[0];

	    status = line_list(&context, &list);
	    if (status) {
		perror(argv[0]);
	    } else {
		status = post_list(&context, bqfx, list);
		if (status) {
		} else {
		    status = find_list(&context, bqfx, list);
		    if (status) {
		    } else {
			status = cast_list(&context, bqfx, list);
			if (status) {
			} else {
			}
		    }
		}

		free_list(list);
	    }

	    excess = x1f4_fini_bqfset(&bqfx);
	    if (excess) {
		perror(argv[0]);

		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}
