/*
 * fs11.c
 * Copyright (C) 2002-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBx1f4l0
# include <libx1f4l0.h>
#endif				/* HAVE_LIBx1f4l0 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
# if !defined HAVE_LIBx1f4l0
# include <xalist.h>
#endif				/* !HAVE_LIBx1f4l0 */

typedef struct context_type {
    int do_storage;
    const char *self;
    void *xa11;
} context_type;

static int case_away(struct context_type *);
static int dump_list(struct context_type *, void *);
static int dump_xa11(struct context_type *);
static int fast_away(struct context_type *);
static int fini_list(struct context_type *, void **);
static int head_line(struct context_type *, void *, char *, unsigned);
static int init_list(struct context_type *, void **);
static int link_node(void *, const char *, unsigned, void *);
static int read_list(struct context_type *, void *);
static int type_text(void *, const char *, unsigned);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    if (context_data->do_storage) {
		status = dump_xa11(context_data);
	    } else {
		status = dump_list(context_data, list);
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    return x1f4_dump_fspost(list, NULL, type_text);
}


static int
dump_xa11(struct context_type *context_data)
{
    unsigned post, size;
    void *xa11;

    xa11 = context_data->xa11;

    x1f4_post_xalist(xa11, &post);
    x1f4_size_xalist(xa11, &size);

    printf(" %7u nodes, %7u bytes\n", post, size);

    return 0;
}


static int
fast_away(struct context_type *context_data)
{
    int status;

    status = x1f4_init_xalist(&context_data->xa11);
    if (status) {
    } else {
	int excess;

	status = case_away(context_data);

	excess = x1f4_fini_xalist(&context_data->xa11);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_fspost(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
head_line(struct context_type *context_data, void *list, char *data,
	  unsigned size)
{
    int status;

    status = x1f4_push_fspost(list, data, size);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_fspost_type fspost;
    struct x1f4_trans_type trans;

    fspost.link = link_node;
    fspost.text = context_data->xa11;
    fspost.trans = &trans;

    trans.context = context_data->xa11;

    trans.free = x1f4_free_xalist;
    trans.link = x1f4_link_xalist;
    trans.mode = x1f4_mode_xalist;

    status = x1f4_init_fspost(list, X1f4_FSPOST_TRANS_MASK, &fspost);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
link_node(void *text, const char *data, unsigned size, void *node)
{
    return x1f4_link_xalist(text, &text, size + sizeof(void *) + 1);
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;

    do {
	char *data;
	int done = 0;
	unsigned size = 0;

	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = head_line(context_data, list, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    }

	    if (done) {
		break;
	    } else {
		while (1) {
		    status = libx1f4l2_read_line
			(context_data->self, &done, &data, &size);
		    if (status) {
			break;
		    }

		    if (!done || size) {
			status = head_line(context_data, list, data, size);
		    }

		    if (1) {
			free(data);
		    }

		    if (status) {
			break;
		    }

		    if (done) {
			break;
		    }
		}
	    }
	}
    } while (0);

    return status;
}


static int
type_text(void *text, const char *data, unsigned size)
{
    fwrite(data, size, 1, stdout);

    return 0;
}


static void
usage(void)
{
    puts("Usage: fs11 [OPTIONS]\n\
Test fs_insert(), leak memory like hell, output some.\n\
\n\
Options:\n\
  -M, --stat-storage		stat list memory storage requirements instea\
d\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.do_storage = 0;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "stat-storage", 0x00,   NULL,   0x4d    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "M", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'M':
		context.do_storage = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = fast_away(&context);
    }

    return status;
}
