/*
 * fs21.c
 * Copyright (C) 2002-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>

#define b__0(byte)			((byte)[0] << 030)
#define b__1(byte)			((byte)[1] << 020)
#define b__2(byte)			((byte)[2] << 010)
#define b__3(byte)			((byte)[3] << 000)

typedef struct context_type {
    const char *self;
    int every_second;
    unsigned line;
} context_type;

static int case_away(struct context_type *);
static int dump_list(struct context_type *, void *);
static int fini_list(struct context_type *, void **);
static int head_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
static int init_list(struct context_type *, void **);
static int miss_half(struct context_type *, void *);
static int read_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    do {
		if (context_data->every_second) {
		    status = miss_half(context_data, list);
		    if (status) {
			break;
		    }
		}

		status = dump_list(context_data, list);
	    } while (0);
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    int status;
    char *line;

    line = (char *) malloc(context_data->line + 1);
    if (!line) {
	status = -1;
    } else {
	unsigned i = 0, size;

	status = 0;

	x1f4_size_fslist(list, &size);
	while (i < size) {
	    unsigned head;
	    unsigned char *data;
	    void *base, *node, *peek, *post, *slip;

	    x1f4_peek_fslist(list, i, &peek, &post, &node, &base);

	    data = peek;

	    head = b__0(data) | b__1(data) | b__2(data) | b__3(data);

	    i++;

	    slip = line;
	    x1f4_type_fslist(node, &slip);
	    strcpy(slip, base);

	    printf("%4u %c <%s>\n", head, *(char *) post, line);
	}

	free(line);
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_fslist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
head_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    int status;
    unsigned offset;
    void *push;

    status = x1f4_push_fslist(list, index, size, data, 1, &offset, &push);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	unsigned char *line;

	line = push;

	line[0] = index >> 030;
	line[1] = index >> 020;
	line[2] = index >> 010;
	line[3] = index >> 000;

	line[offset] = index & 1 ? '+' : '*';

	if (context_data->line < size) {
	    context_data->line = size;
	}
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_fslist_type fslist;

    fslist.base = 4;

    status = x1f4_init_fslist(list, 0, &fslist);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
	context_data->line = 0;
    }

    return status;
}


static int
miss_half(struct context_type *context_data, void *list)
{
    int status;
    unsigned i = 1, size;

    status = 0;

    x1f4_size_fslist(list, &size);
    while (i < size) {
	status = x1f4_miss_fslist(list, i);
	if (status) {
	    break;
	} else {
	    i++;
	    size--;
	}
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = head_line(context_data, list, i, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    } else {
		i++;
	    }

	    if (done) {
		break;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: fs21 [OPTIONS]\n\
Test fs_insert(), output input.\n\
\n\
Options:\n\
  -2, --second			delete every second line before output\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.every_second = 0;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "second",	    0x00,   NULL,   0x32    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "2", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case '2':
		context.every_second = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
