/*
 * ls12.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <inter.h>
#include <libx1f4l2.h>
#include <lsas.h>

typedef struct context_type {
    const char *self;
    unsigned class, frame;
} context_type;

static int call_list(struct context_type *, void *, unsigned);
static int call_node(void *, void *, const char *, unsigned);
static int case_away(struct context_type *);
static int fast_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int read_list(struct context_type *, void *, unsigned *);

static void usage(void);

static int
call_list(struct context_type *context_data, void *list, unsigned copy)
{
    return x1f4_span_lsnote(list, (char *) context_data->self, call_node);
}


static int
call_node(void *text, void *data, const char *node, unsigned size)
{
    unsigned call = 0;
    const unsigned char *slip;

    slip = (const void *) node;

    slip += size;
    if (4 < size) {
	size = 4;
    }
    while (size) {
	call <<= 8;
	call |= *--slip;
	size--;
    }

    call ^= *(unsigned *) data;

    if (call) {
	fprintf(stderr, "%s: cannot match `%s' record\n", (char *) text, node);
    }

    return call;
}


static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	unsigned copy;

	status = read_list(context_data, list, &copy);
	if (status) {
	} else {
	    status = call_list(context_data, list, copy);
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
fast_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    int status;
    void *fail;

#if __BILL_FOR_NOTE__
    status = x1f4_post_lsbill(list, data, size, sizeof(unsigned), &fail);
#else
    status = x1f4_post_lsnote(list, data, sizeof(unsigned), &fail);
#endif				/* __BILL_FOR_NOTE__ */
    if (status == X1f4_LSNOTE_EVER_MATCH) {
	fprintf(stderr, "%s: duplicate `%s' line\n", context_data->self, data);
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	    unsigned call = 0, *host;
	    unsigned char *slip;

	    slip = (void *) data;

	    host = fail;

	    slip += size;
	    if (4 < size) {
		size = 4;
	    }
	    while (size) {
		call <<= 8;
		call |= *--slip;
		size--;
	    }

	    *host = call;
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_lsnote(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_lsnote_type lsnote;

#if __FIX_INTERNAL__
    lsnote.inner = context_data->class;
#endif				/* __FIX_INTERNAL__ */
#if __FIX_EXTERNAL__
    lsnote.outer = context_data->frame;
#endif				/* __FIX_EXTERNAL__ */

    status = x1f4_init_lsnote
	(list, X1f4_LSNOTE_INNER_MASK | X1f4_LSNOTE_OUTER_MASK, &lsnote);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list, unsigned *copy)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, miss = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (done) {
	    } else {
#if __MISS_VOID_LINES__
		if (size) {
#endif				/* __MISS_VOID_LINES__ */
		    status = fast_line(context_data, list, i, data, size);
		    if (miss < size) {
			miss = size;
		    }
#if __MISS_VOID_LINES__
		}
#endif				/* __MISS_VOID_LINES__ */
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    } else {
		i++;
	    }

	    if (done) {
		break;
	    }
	}
    }

    if (status) {
    } else {
	*copy = miss;
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: ls12 [OPTIONS]\n\
Test " lsname "_insert(), output sorted input.\n\
\n\
Options:\n\
  -e, --external L		set external nodes size as L bytes\n\
  -i, --internal L		set internal nodes size as L bytes\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.class = 256;
    context.frame = 256;

    {
	const char *class_specs = NULL, *frame_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "external",	    0x01,   NULL,   0x65    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   "internal",	    0x01,   NULL,   0x69    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "e:i:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'e':
		frame_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'i':
		class_specs = optarg;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (class_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.class, class_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse internal node width specif"
			"ication: `%s'\n", argv[0], class_specs);

		return 1;
	    } else {
		if (context.class < 4) {
		    fprintf(stderr, "%s: will not use internal node widths l"
			    "ess than 4 (like the indicated 1 << %u)\n",
			    argv[0], context.class);

		    return 1;
		}
	    }
	}

	if (frame_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.frame, frame_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse external node width specif"
			"ication: `%s'\n", argv[0], frame_specs);

		return 1;
	    } else {
		if (context.frame < 1) {
		    fprintf(stderr, "%s: will not use external node widths l"
			    "ess than 4 (like the indicated %u << 2)\n",
			    argv[0], context.frame);

		    return 1;
		}
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
