/*
 * sx40.c
 * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, Ciprian Niculescu
 * Copyright (C) 2009, 2010, 2011, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <count2.h>
#include <libx1f4l2.h>
#include <s3xx.h>

#define LX				10

#define VX				1048576

typedef struct context_type {
    const char *self;
    int do_profile, do_size;
    unsigned count;
} context_type;

static int case_list(struct context_type *, void *, void *);
static int free_list(void *);
static int link_list(struct context_type *, void **);
static int pick_node(void *, const void *);
static int size_list(struct context_type *, void *);

static void usage(void);

static const char digits[] = "0123456789";

static int
case_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int status;
    unsigned i;

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	const void *data;

	i--;
	status = x1f4_find_s3deck(bqfx, list, &data);
	if (status) {
	    status = x1f4_pass_s3deck(bqfx, list);
	    if (status) {
		fprintf(stderr, "%s: cannot delete data\n",
			context_data->self);
		if (1) {
		    break;
		}
	    }
	} else {
	    status = x1f4_post_s3deck(bqfx, list, NULL);
	    if (status) {
		fprintf(stderr, "%s: cannot insert data\n",
			context_data->self);
		if (1) {
		    break;
		}
	    }
	}

	list = (char *) list + 1 + LX;
    }

    if (i) {
	status = 1;
    } else {
	status = 0;

	if (context_data->do_profile) {
	    finish_count2_l("", 55);
	}
    }

    return status;
}


static int
free_list(void *list)
{
    free(list);

    return 0;
}


static int
link_list(struct context_type *context_data, void **link)
{
    int status;
    unsigned size;
    void *list;

    size = context_data->count;

    list = malloc((LX + 1) * size);
    if (list) {
	unsigned i;

	srand48(11);

	status = 0;

	*link = list;

	i = size;
	while (i) {
	    char *o;
	    unsigned e;

	    i--;

	    e = (unsigned) (size * drand48() / 4) * 271828183u;

	    o = (char *) list + LX;

	    *o = 0;

	    do {
		--o;
		*o = digits[e % 10];
		e /= 10;
	    } while (e);
	    while (o != list) {
		--o;
		*o = 32;
	    }

	    list = (char *) list + 1 + LX;
	}
    } else {
	status = -1;
    }

    return status;
}


static int
pick_node(void *text, const void *data)
{
    unsigned *c;

    c = text;
    (*c)++;

    return 0;
}


static int
size_list(struct context_type *context_data, void *bqfx)
{
    int status;

    if (context_data->do_size) {
	unsigned count = 0;

	status = x1f4_lime_s3deck(bqfx, &count, pick_node);
	if (status) {
	} else {
	    printf("%u\n", count);
	}
    } else {
	status = 0;
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: sx40 [OPTIONS]\n\
Test the " sxname " associative array.\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -p, --profile			time operations\n\
  -s				print the array final size\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status;
    struct context_type context;

    context.do_profile = 0;
    context.do_size = 0;

    context.count = VX;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *size_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "profile",	    0x00,   NULL,   0x70    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:ps", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'p':
		context.do_profile = 1;
		break;
	    case 's':
		context.do_size = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse test set width specificati"
			"on: `%s'\n", argv[0], size_specs);

		return 1;
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	void *bqfx;

	status = x1f4_init_s3deck(&bqfx, 0, NULL);
	if (status) {
	    perror(argv[0]);
	} else {
	    int excess;
	    void *list;

	    context.self = argv[0];

	    status = link_list(&context, &list);
	    if (status) {
		perror(argv[0]);
	    } else {
		status = case_list(&context, bqfx, list);
		if (status) {
		} else {
		    status = size_list(&context, bqfx);
		}

		free_list(list);
	    }

	    excess = x1f4_fini_s3deck(&bqfx);
	    if (excess) {
		perror(argv[0]);

		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}
